-- AngelPlayer: Battlemod support
-- Enables a Chaos Energy powered 'Howling' to better fight against enemies with more elaborate behaviours.

local battoru_set
rawset(_G, "AP_BattleMod", function()

	local b = CBW_Battle
	if not b then return end		-- Mod not loaded
	if battoru_set then return end	-- Mod already loaded in

	b.SkinVars["kanade"] = {

		flags = SKINVARS_GUARD,
		weight = 100,
		shields = 0,	-- Offence axed.

		special =	function(mo, act)

						local p = mo.player

						p.actiontext = "Guard Skill: Howling"
						p.actionrings = 20
						p.actionsuper = true	-- getting hurt doesn't cancel the action.

						-- (unique guard frame in AP_Extras)
						-- Hand Sonic priority:
						if mo.ap_handstate == HS_DASH
							CBW_Battle.SetPriority(p,0,0,"knuckles_glide",3,3,"Hand Sonic")
							-- Use Knuckles glide check / with different priorities. Don't question it.
						end


						-- button press garbage
						if not p.actionstate and act == 1
						--and P_IsObjectOnGround(mo)
							p.actionstate = 1
							p.actiontime = 1

							b.PayRings(p)
						end
						if p.actionstate == 1		-- Howling init
							-- readying attack:
							mo.state = S_PLAY_HOWLING_PREP1
							p.powers[pw_nocontrol] = 2
							p.actionstate = 2

						elseif p.actionstate == 2	-- Howling prep anim
							p.powers[pw_nocontrol] = 2
							p.actiontext = "Hold to keep firing"

							p.actiontime = $+1
							if mo.state != S_PLAY_HOWLING_PREP1
							and mo.state != S_PLAY_HOWLING_PREP2	-- bruh
							and mo.state != S_PLAY_HOWLING
								mo.state = S_PLAY_HOWLING_PREP2
								mo.tics = max(1, $-p.actiontime)	-- mercy mechanic on this
							end


							if mo.state == S_PLAY_HOWLING	-- prep finished
								p.actionstate = 3
								p.actiontime = 1	-- reset actiontime.
								P_FlashPal(p, 1, 1)
								if p == consoleplayer
									P_StartQuake(FRACUNIT*64, 6)
								end
								S_StartSound(mo, sfx_howl1)
							end

						elseif p.actionstate == 3	-- Howling proper!
							p.actiontext = "Hold to keep firing"
							p.actiontime = $+1
							p.powers[pw_nocontrol] = 3

							if ((p.actiontime) % (TICRATE/4) == 0)
								P_GivePlayerRings(p, -1)
							end

							p.drawangle = mo.angle	-- keep updating this.

							if leveltime%10 == 0
							and p.actiontime > 5
								S_StartSound(mo, sfx_howl2)
							end

							if p == consoleplayer
							and p.actiontime > 5
								P_StartQuake(FRACUNIT*20, 6)
							end

							if leveltime%5 == 0
								mo.height = $*3	-- help missile spawning.
								local m = P_SpawnPlayerMissile(mo, MT_HOWLING)
								mo.height = $/3
								if m and m.valid
									m.color = mo.color
									m.angle = $ + ANG1*90
									m.scale = $*2
									m.destscale = mo.scale*8
									m.scalespeed = mo.scale/12
								end
							end

							local actp = p.cmd.buttons & BT_ATTACK

							-- give 0.5 sec of margin to use the ability with only 20 rings.
							if (not p.rings and p.actiontime > TICRATE*3/2)
							or (not actp and p.actiontime > TICRATE*3/2)

								b.ApplyCooldown(p, TICRATE*30)
								p.actiontime = 0
								p.actionstate = 0
								mo.state = S_PLAY_STND
								p.powers[pw_nocontrol] = 0
							end
						end

					end,

	}

	battoru_set = true
	print("AP: 'BattleMod' library loaded succesfully. 'Howling' is now available in Battle modes.")
end)