-- General definitions

/*
	GENERAL CONSTANTS
*/

-- ability constants
rawset(_G, "CA_ANGELWINGS", 55)
rawset(_G, "CA2_HANDSONIC", 55)

-- arbitrary momz to apply midfall. AP_MIDFALLTHRESHOLD <= momz <= 0
rawset(_G, "AP_MIDFALLTHRESHOLD", -2)

-- arbitrary momz to apply when using CA_ANGELWINGS
rawset(_G, "AP_ANGELWINGSTHRUST", 17)
rawset(_G, "AP_ANGELWINGSSLOWFALL", 4)

-- stuff for the updraft
rawset(_G, "AP_MAXUPDRAFTSPEED", 12)

-- minimum speed for CA2_HANDSONIC
rawset(_G, "AP_HANDSONICMINSPEED", 25)

-- Stuff for Overdrive:
rawset(_G, "AP_OVERDRIVE_MAXCHARGE", 100)	-- max charge for overdrive
rawset(_G, "AP_OVERDRIVE_KILLCHARGE", 15)	-- charge gained / kill
rawset(_G, "AP_OVERDRIVE_DURATION", 20)	-- 20s

/*
	SOUND DEFINITIONS
*/
freeslot("sfx_wngfl", "sfx_wngapr", "sfx_hsdash", "sfx_hsslas", "sfx_skew")
sfxinfo[sfx_wngfl].caption = "Takeoff"
sfxinfo[sfx_wngapr].caption = "Appearing Wings"
sfxinfo[sfx_hsdash].caption = "Dash"
sfxinfo[sfx_hsslas].caption = "Slash"
sfxinfo[sfx_cdfm28].caption = "Bump"
sfxinfo[sfx_skew].caption = "Skewer"

/*
	SPR2 DEFINITIONS
*/

-- Mid-Fall frame: mostly to smoothen hair from spring to fall
-- spr2 for spring midfall
freeslot("SPR2_MFAL", "S_PLAY_MIDFALL")
spr2defaults[SPR2_MFAL] = SPR2_FALL

states[S_PLAY_MIDFALL] = {SPR_PLAY, SPR2_MFAL, 4, nil, 0, 0, S_PLAY_MIDFALL}

-- spr2 for float start / end
freeslot("SPR2_FLTE", "S_PLAY_FLOATEND", "SPR2_FLTS", "S_PLAY_FLOATSTART")
spr2defaults[SPR2_FLTE] = SPR2_MFAL
spr2defaults[SPR2_FLTS] = SPR2_FLTE

states[S_PLAY_FLOATSTART] = {SPR_PLAY, SPR2_FLTS, 4, nil, 0, 0, S_PLAY_FLOAT}
states[S_PLAY_FLOATEND] = {SPR_PLAY, SPR2_FLTE, 4, nil, 0, 0, S_PLAY_FLOATEND}

-- spr2 for guard frame
freeslot("SPR2_GARD", "S_PLAY_GUARD")
states[S_PLAY_GUARD] = {SPR_PLAY, SPR2_GARD, 2, nil, 0, 0, S_PLAY_GUARD}

-- spr2 for hand sonic
freeslot("SPR2_HWLP", "SPR2_HOWL", "S_PLAY_HOWLING_PREP1", "S_PLAY_HOWLING_PREP2", "S_PLAY_HOWLING")
states[S_PLAY_HOWLING_PREP1] = {SPR_PLAY, SPR2_GARD, 10, nil, 0, 0, S_PLAY_HOWLING_PREP2}
states[S_PLAY_HOWLING_PREP2] = {SPR_PLAY, SPR2_HWLP, TICRATE, nil, 0, 0, S_PLAY_HOWLING}

states[S_PLAY_HOWLING] = {SPR_PLAY, SPR2_HOWL, 2, nil, 0, 0, S_PLAY_HOWLING}

-- Hand Sonic (uses SPR2_MLEE)
freeslot("S_PLAY_HANDSONIC")
states[S_PLAY_HANDSONIC] = {SPR_PLAY, SPR2_MLEE, 4, nil, 0, 0, S_PLAY_HANDSONIC}

-- Hugs are cool and wholesome
freeslot("SPR2_BLSH", "SPR2_KHUG", "S_KANADE_B", "S_KANADE_H")
states[S_KANADE_B] = {SPR_PLAY, SPR2_BLSH, 1, nil, 0, 0, S_KANADE_B}
states[S_KANADE_H] = {SPR_PLAY, SPR2_KHUG, 1, nil, 0, 0, S_KANADE_H}

/*
	WINGS OBJECT
*/
freeslot("MT_KANADEWING", "S_KANADEWING", "SPR_KWNG")
mobjinfo[MT_KANADEWING] = {

	doomnednum = -1,
	spawnstate = S_KANADEWING,
	spawnhealth = 1000,
	radius = 48<<FRACBITS,
	height = 48<<FRACBITS,
	flags = MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY
}
states[S_KANADEWING] = {SPR_KWNG, FF_PAPERSPRITE, -1, nil, 0, 0, S_KANADEWING}

/*
	UPDRAFT (Created by using Angel Wings, can be used by other players if they hold jump in it)

*/
freeslot("MT_KANADEUPDRAFT", "SPR_UDRT")
mobjinfo[MT_KANADEUPDRAFT] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	radius = 127 << FRACBITS,
	height = 48 << FRACBITS,	-- in actuality, much taller, but we handle this in mobjcollide.
	flags = MF_NOCLIPHEIGHT|MF_NOGRAVITY,
}

-- Could be handled in a MobjCollide hook but this is more convenient to handle it here.
addHook("MobjThinker", function(mo)

	-- functionality.

	if not mo.target or not mo.target.valid
		P_RemoveMobj(mo)
		return
	end	-- don't do anything if our target ceases to exist, just desist.

	for p in players.iterate do
		local pmo = p.mo
		if not pmo or not pmo.valid continue end	-- don't interact with spectators or whoever spawned it.
		if not AP_friendlyPlayer(mo.target.player, p) continue end

		if p.cmd.buttons & BT_JUMP	-- holding jump
		and R_PointToDist2(mo.x, mo.y, pmo.x, pmo.y) < mo.scale*192	-- 192 radius circle AOE. (check after jump for efficiency)
		and (pmo.eflags & MFE_VERTICALFLIP) == (mo.eflags & MFE_VERTICALFLIP)
		and (pmo.eflags & MFE_UNDERWATER) == (mo.eflags & MFE_UNDERWATER)	-- yes this is a rather lazy check
		--and not (pmo.ap_wingstate == WS_PREP)
		and not (pmo.ap_handstate == HS_DASH)	-- slight exceptions to avoid kanade herself being caught by weird jank

		-- simplify our life, make sure flipped flags are matching on both ends.

			-- special case: for angel wings, only apply thrust in slowfall state.
			--print(pmo.ap_wingstate)
			if pmo.ap_wingstate
			and pmo.ap_wingstate ~= WS_FLOAT
				continue	-- no
			end

			-- make sure no FOFs are in the way
			-- and that we're at the proper height and whatnot (128 above the updraft origin)

			if pmo.eflags & MFE_VERTICALFLIP

				if pmo.z > mo.z					-- player under the updraft
				or pmo.z < mo.z - 300*mo.scale	-- player too far above the upraft
				or pmo.ceilingz <= mo.floorz	-- fof in the way
				or pmo.momz < -AP_MAXUPDRAFTSPEED*mo.scale	-- maximum momentum reached
					continue	-- nope
				end

			else
				if pmo.z < mo.z					-- player under the updraft
				or pmo.z > mo.z + 300*mo.scale	-- player too far above the updraft
				or pmo.floorz >= mo.ceilingz	-- fof in the way
				or pmo.momz > AP_MAXUPDRAFTSPEED*mo.scale	-- maximum momentum reached
					continue	-- nope
				end
			end

			pmo.momz = $ + gravity*4 * P_MobjFlip(pmo) / (mo.eflags & MFE_UNDERWATER and 2 or 1)
		end
	end

	//print((mo.watertop/FRACUNIT).."|"..(mo.z/FRACUNIT).."|"..(mo.eflags & MFE_UNDERWATER)))

	-- visuals:
	for i = 1, 2 do
		local x = mo.x + P_RandomRange(-128, 128)*mo.scale
		local y = mo.y + P_RandomRange(-128, 128)*mo.scale

		local d = P_SpawnMobj(x, y, mo.z, mo.eflags & MFE_UNDERWATER and MT_SMALLBUBBLE or MT_THOK)
		d.eflags = mo.eflags & MFE_VERTICALFLIP
		d.tics = -1
		d.fuse = TICRATE/2
		d.scale = mo.scale*3/2
		d.scalespeed = mo.scale/36
		P_SetObjectMomZ(d, AP_MAXUPDRAFTSPEED*FRACUNIT*2)

		d.sprite = i & 1 and SPR_ADST or SPR_UDRT
		if mo.eflags & MFE_UNDERWATER
			if i & 1
				d.state = S_SPINDUST_BUBBLE1
			end
			d.momz = $/2
		end

		d.frame = FF_TRANS60
		d.colorized = true
		d.color = SKINCOLOR_WHITE

		-- Don't display the gusts if you aren't friendly to her.
		if not AP_friendlyPlayer(mo.target.player, consoleplayer)
			d.flags2 = $|MF2_DONTDRAW
		end	-- Not my problem!
	end

end, MT_KANADEUPDRAFT)


/*
	HANDSONIC
*/
-- Hand Sonic hitbox detection
freeslot("MT_HS_HITBOX")
mobjinfo[MT_HS_HITBOX] = {
	doomednum = -1,
	spawnstate = S_INVISIBLE,
	radius = 64<<FRACBITS,
	height = 64<<FRACBITS,
	flags = MF_NOCLIPHEIGHT|MF_NOGRAVITY
}

freeslot("SPR_HSSL")
-- Hand Sonic reach slash
for i = 1, 3
	freeslot("S_HANDSONIC_SLASH"..i)
end
for i = 0, 2
	states[S_HANDSONIC_SLASH1+i] = {SPR_HSSL, i, 1, nil, 0, 0, i < 2 and S_HANDSONIC_SLASH1+(i+1) or S_NULL}
end

-- Stolen from srb2p. lmao
freeslot("S_HANDSONIC_STAB", "SPR_HSST")
function A_StabEnd(a)
	if not a or not a.valid return end

	local mo = P_SpawnMobj(a.x, a.y, a.z, MT_THOK)
	mo.tics = TICRATE/8
	mo.color = a.color
	mo.scale = a.scale/2
	mo.destscale = a.scale*4
	mo.scalespeed = a.scale/2
	mo.sprite = SPR_HSST
	mo.frame = A|FF_FULLBRIGHT
	mo.momx = a.momx
	mo.momy = a.momy

	mo = P_SpawnMobj(a.x, a.y, a.z, MT_THOK)
	mo.tics = TICRATE/4
	mo.fuse = mo.tics
	mo.color = a.color
	mo.scale = a.scale/8
	mo.destscale = a.scale*8
	mo.scalespeed = a.scale/2
	mo.sprite = SPR_HSST
	mo.frame = A|FF_FULLBRIGHT
	mo.momx = a.momx
	mo.momy = a.momy
end

states[S_HANDSONIC_STAB] = {SPR_NULL, A|FF_FULLBRIGHT, 1, A_StabEnd, 0, 0, S_NULL}

-- some states for visual effects:
freeslot("SPR_WSRK")
for i = 1, 5
	freeslot("S_WINGSPARK"..i)
end

for i = 0, 4
	states[S_WINGSPARK1+i] = {SPR_WSRK, i|FF_FULLBRIGHT, 4, nil, 0, 0, i < 4 and S_WINGSPARK1+(i+1) or S_NULL}
end

/*
	HOWLING
*/

freeslot("sfx_howl1", "sfx_howl2")
freeslot("MT_HOWLING", "SPR_HOWL", "S_KHOWL1", "S_KHOWL2", "S_KHOWL3")
local tlookup = {TR_TRANS40, TR_TRANS60, TR_TRANS70}
for i = 0, 2 do
	states[S_KHOWL1+i] = {SPR_HOWL, i|tlookup[i+1]|FF_PAPERSPRITE, 10, nil, 0, 0, i < 2 and S_KHOWL2+i or S_NULL}
end

mobjinfo[MT_HOWLING] = {
	doomednum = -1,
	spawnhealth = 1000,
	spawnstate = S_KHOWL1,
	speed = 64 << FRACBITS,
	radius = 32 << FRACBITS,
	height = 48 << FRACBITS,
	flags = MF_NOGRAVITY|MF_MISSILE,
}

local function howlingCollide(mo, mo2)

	if not mo or not mo.valid or not mo2 or not mo2.valid return false end
	if not mo2.player return false end
	local source = mo.target
	if not source or not source.valid return false end	-- !?
	if mo2 == source return false end
	if mo2.player.powers[pw_flashing] return false end	-- don't bother with this.

	mo.playerhits = $ or {}

	-- check if we hit mo2 already
	for i = 1, #mo.playerhits
		if mo.playerhits[i] == mo2
			return false
		end	-- leave him alone. for now.
	end

	if not AP_friendlyPlayer(source.player, mo2.player)
		-- *Gasp* The enemy!
		if mo2.player.rings <= 5
			P_DamageMobj(mo2, mo, source)
		else
			P_GivePlayerRings(mo2.player, -5)
			P_PlayRinglossSound(mo2, mo2.player)

			source.player.score = $+5
		end
		mo.playerhits[#mo.playerhits+1] = mo2
	end

	return false
end

-- Still don't know how this shit works and I've been coding for this game for about 6 years now lol
addHook("MobjCollide", howlingCollide, MT_HOWLING)
addHook("MobjMoveCollide", howlingCollide, MT_HOWLING)

/*
	GENERALIST FUNCTIONS
*/

rawset(_G, "AP_advanceGhost", function(mo)
	assert(mo, "AP_advanceGhost: no valid mo???")

	if leveltime & 1 return end
	local g = P_SpawnGhostMobj(mo)
	g.colorized = true
	g.rollangle = mo.rollangle

	return g
end)

-- silv_blockingWall:
-- wall busting
local function blockbusters(mo ,bustsec)
	for rover in bustsec.ffloors() -- search through FOFs in sector
		if not (rover.flags & FF_EXISTS) continue end -- FOF does not "exist" in sector!
		if not (rover.flags & FF_BUSTUP) continue end -- FOF is not bustable, ignore it

		if mo.z+mo.momz > rover.topheight -- too high
			continue
		end
		if mo.z + mo.momz+mo.height < rover.bottomheight -- too low
			continue
		end

		EV_CrumbleChain(rover) -- DESTROY IT
		return true
	end
end

-- mo will break bustable blocks in its path
rawset(_G,"AP_FOFBuster", function(mo)
	local secs = {}

	secs[1] = R_PointInSubsector(mo.x, mo.y).sector
	secs[2] = R_PointInSubsector(mo.x+(mo.momx/2), mo.y + (mo.momy/2)).sector
	secs[3] = R_PointInSubsector(mo.x+(mo.momx), mo.y + (mo.momy)).sector
	secs[4] = R_PointInSubsector(mo.x+(mo.momx*2), mo.y + (mo.momy*2)).sector
	for i = 1, #secs
		local sec = secs[i]
		if blockbusters(mo, sec)
			return true	-- busted sector
		end
	end
end)

-- kills a "pain mobj"
freeslot("MT_AP_PAINMOBJ", "S_AP_PAINMOBJ")
mobjinfo[MT_AP_PAINMOBJ] = {
	doomednum = -1,
	spawnstate = S_AP_PAINMOBJ,
	radius = FRACUNIT,
	height = FRACUNIT,
	flags = MF_NOCLIPHEIGHT,
	spawnhealth = 1000,
}
states[S_AP_PAINMOBJ] = {SPR_NULL, A, -1, nil, 0, 0, S_AP_PAINMOBJ}
addHook("MobjThinker", function(mo)
	if leveltime & 1
		mo.flags2 = $|MF2_DONTDRAW
	else
		mo.flags2 = $ & ~MF2_DONTDRAW
	end
end, MT_AP_PAINMOBJ)

rawset(_G, "AP_killPainMobj", function(mo)

	-- check if mo is related to an ENEMY:
	if mo.target and mo.target.valid and mo.target.flags & MF_ENEMY	return end	--nope
	if mo.tracer and mo.tracer.valid and mo.tracer.flags & MF_ENEMY return end	--nope either
	if mo.flags & MF_FIRE return end	-- can't break fire now can ya...

	local curobj = mo
	local s = 1
	while curobj and curobj.valid	-- get out if curobj isn't valid somhehow

		local m = P_SpawnMobjFromMobj(curobj, 0, 0, 0, MT_AP_PAINMOBJ)
		m.height = curobj.height
		m.radius = curobj.radius
		m.sprite = curobj.sprite
		m.frame = curobj.frame

		m.momx = P_RandomRange(-8, 8)*curobj.scale
		m.momy = P_RandomRange(-8, 8)*curobj.scale
		m.momz = 8*curobj.scale*P_MobjFlip(curobj)

		if s
			S_StartSound(m, sfx_s3k59)
			s = nil	-- don't earrape pls
		end
		m.fuse = TICRATE
		local buf = curobj
		curobj = curobj.hprev or curobj.hnext	-- whatever works
		P_RemoveMobj(buf)	-- lol die
	end
end)

-- return true if p and p2 are "friends"
rawset(_G, "AP_friendlyPlayer", function(p, p2)

	if p == p2
		return true	-- yeah I guess I'm my own friend?
	end

	if gametyperules & GTR_FRIENDLY
		return true	-- always return true in that case
	end

	if gametyperules & GTR_TEAMS
		return p.ctfteam == p2.ctfteam	-- same team
	end

	if gametyperules & GTR_TAG
		return p.pflags & PF_TAGIT == p2.pflags & PF_TAGIT	-- tagged/untagged
	end

	return false	-- nope
end)

-- colourcodes a player's name to match their team
rawset(_G, "AP_teamName", function(p)

	if not (gametyperules & GTR_TEAMS)
		return p.name
	else
		return (p.ctfteam == 1 and "\x85" or "\x84")..p.name.."\x80"	-- spectators won't get hurt so this ternary will do.
	end
end)

-- are we in battle mode...?
-- stg couldn't find any other clean future proof way to do this :mike:
rawset(_G, "AP_battleMode", function()
	local b = CBW_Battle
	if not b return false end	-- definitely not battle.

	for k, v in pairs(b.Gametypes.TeamScoreType)	-- whatever works
		if gametype == k
			return true
		end
	end

	-- yes this function is fucking stupid.
	return false
end)

