-- AngepPlayer: Handler for CA2_HANDSONIC
-- Enables an offensive slashing move. Will overpower any direct contact damage and impale any touched enemy to slice them in half.
-- To compensate for the lack of range, an additional slice is performed at the end of the dash.
-- Can be chained with Angel Wings to divebomb and destroy enemies in a radius at the impact point.

local spinbutt = BT_SPIN or BT_USE

-- Hack for spikes.
mobjinfo[MT_SPIKE].flags = $ & ~MF_NOBLOCKMAP		-- WHY IS THIS HERE.
mobjinfo[MT_WALLSPIKE].flags = $ & ~MF_NOBLOCKMAP	-- WHY IS THIS HERE X2.

-- Quick constants for cleanliness
local hs_constants = {
	"HS_NONE",
	"HS_DASH",
	"HS_DIVE",
	"HS_FREEFALL",
	"HS_DONE",
}
for i = 1, #hs_constants do
	rawset(_G, hs_constants[i], i-1)
end

-- Ability trigger
addHook("SpinSpecial", function(p)
	local mo = p.mo
	if not p.mo return end	--?
	if p.charability2 != CA2_HANDSONIC return end

	if (not mo.ap_handstate or mo.ap_handstate == HS_FREEFALL)
	and not (mo.ap_wingstate == WS_PREP)
	and not (mo.ap_wingstate == WS_UP)
	--and not (mo.ap_wingstate == WS_FLOAT)	-- allow usage while floating, it will override it however
	and not P_PlayerInPain(p)
	and not p.powers[pw_carry]
	and mo.health
	and not mo.ap_hangbuf	-- we need to be able to drop from hang ropes
	and mo.ap_spinbuf == 1	-- and finally, make sure this is the first tic the bytton is being pressed for...
	and not p.exiting		-- not after level complete!

	and not (gametyperules & GTR_RACE and leveltime < TICRATE*4)	-- dear god leave me alone......

	and not p.gotflag

		mo.ap_hsjumpbuf = nil	-- remove jump buffer

		if (not mo.ap_wingstate and mo.ap_handstate ~= HS_FREEFALL)
		or mo.ap_overdrive	-- overdrive has infinite dashes
			mo.ap_handstate = HS_DASH
			mo.ap_handtimer = TICRATE/4		-- dash duration
		else
			mo.ap_handstate = HS_DIVE
			mo.ap_handtimer = 1
		end

		-- save current angle
		mo.ap_hsaveangle = mo.angle
		p.ap_hsavestats = {p.normalspeed, p.thrustfactor, p.jumpfactor}
		p.normalspeed, p.thrustfactor, p.jumpfactor = 0,0,0

		-- get current speed:
		local spd = R_PointToDist2(0, 0, mo.momx, mo.momy)	-- + 10*mo.scale
		local normalspd = p.ap_hsavestats[1] * ((p.powers[pw_sneakers] or p.powers[pw_super]) and 2 or 1)
		if spd < FixedMul(mo.scale, normalspd*3/2)
			spd = $ + 10*mo.scale	-- add 10 spd if below normalspeed*1.5, otherwise, keep speed.
		end

		local minspeed = AP_HANDSONICMINSPEED
		if mo.ap_overdrive
			minspeed = p.ap_hsavestats[1]/FRACUNIT *2/3	-- be nice to players! :)
		end
		if mo.ap_wingstate	-- high base dive speed
			minspeed = $*3/2
		end

		spd = max($, minspeed*mo.scale)
		mo.ap_hsavespeed = spd

		mo.state = S_PLAY_HANDSONIC

		S_StartSound(mo, sfx_hsdash)
	end
end)

local function AP_hresetStats(p)
	if not p.ap_hsavestats or not #p.ap_hsavestats return end

	p.normalspeed = p.ap_hsavestats[1] or skins[p.realmo.skin].normalspeed
	p.thrustfactor = p.ap_hsavestats[2] or skins[p.realmo.skin].thrustfactor
	p.jumpfactor = p.ap_hsavestats[3] or skins[p.realmo.skin].jumpfactor
	p.ap_hsavestats = nil
end

-- Make sure to reset everything on mapload:
addHook("MapLoad", do
	for p in players.iterate do
		AP_hresetStats(p)
	end
end)

-- Scan for enemies in front of mo, if there are enemies, damages them and spawns a slash GFX
local function AP_handSonicSlash(mo, range)

	local p = mo.player
	local a = mo.angle
	local dist = mo.scale*96	-- 64 radius box

	-- kill skewed enemies:
	if mo.ap_skewer and #mo.ap_skewer
		for i = 1, #mo.ap_skewer
			local m = mo.ap_skewer[i]
			if not m or not m.valid continue end	-- will be emptied
			m.flags = m.info.flags | MF_SHOOTABLE
			AP_sliceBadnik(m, mo)
			--AP_addOverdrive(p)
			P_DamageMobj(m, mo)
		end
	end

	mo.ap_skewer = nil

	local m = P_SpawnMobjFromMobj(mo, FixedMul(dist, cos(a)), FixedMul(dist, sin(a)), 0, MT_HS_HITBOX)
	-- scan blockmap on the box:
	searchBlockmap("objects", function(ref, found)

		if found ~= mo and found ~= ref	-- don't damage ourselves or the helper, idiot.
		and abs(mo.z - found.z) <= mo.scale*64

			if range
				if R_PointToDist2(mo.x, mo.y, found.x, found.y) > range
					return
				end
			end

			if found and found.valid and found.health and (found.flags & MF_ENEMY|MF_MONITOR|MF_BOSS or found.flags & MF_PAIN)	-- this works too
			and (found.flags & MF_SHOOTABLE or found.type == MT_SPIKE or found.type == MT_WALLSPIKE)
			and not (found.flags2 & MF2_FRET)	-- multi hit enemy being damaged, ignore it.

			or found.type == MT_PLAYER
			and found.player
			and not AP_friendlyPlayer(mo.player, found.player)

				if found.flags & MF_PAIN
				and not found.player	-- failsafe JUST IN CASE someone tries something stupid.
					AP_killPainMobj(found)
				end

				if found.type == MT_SPIKE
				or found.type == MT_WALLSPIKE
					P_KillMobj(found, mo, mo)	-- I am suffering.
				else
					if found.health == 1	-- precisely 1 HP
						AP_sliceBadnik(found, mo)
						--AP_addOverdrive(p)
					end

					-- Battlemod: Players with 1def and above are only knocked back by the slice:
					if found.player
					and not found.player.powers[pw_flashing]
					and AP_battleMode()

						if found.player.battle_def
							-- don't take damage
							S_StartSound(found, sfx_spkdth)
							found.player.pflags = $ & ~(PF_SPINNING)
							found.state = S_PLAY_FALL
							found.momx = mo.momx
							found.momy = mo.momy
							found.momz = P_MobjFlip(mo)*10*mo.scale
							-- fly away

							return	-- but go off damage free!
						end
					end

					P_DamageMobj(found, mo, mo)
				end
			end
		end
	end, m)

	S_StartSound(mo, sfx_hsslas)
	local s = P_SpawnMobjFromMobj(m, 0, 0, 0, MT_THOK)
	s.state = S_HANDSONIC_SLASH1
	s.color = mo.color
	s.scale = mo.scale*3/2

	P_RemoveMobj(m)	-- you've outlasted your usefulness, to the death dimension you go
end

-- Ability thinker
rawset(_G, "AP_handSonic", function(p)
	local mo = p.mo

	if mo.ap_hangbuf
		mo.ap_hangbuf = $-1
	end

	if p.powers[pw_carry]
	or P_PlayerInPain(p)
		if p.powers[pw_carry]
			mo.ap_hangbuf = 2	-- make sure hanging doesn't let us trigger
		end
		AP_hresetStats(p)
		mo.ap_handstate = HS_NONE
	end

	-- handle skewer
	if mo.ap_skewer
	and #mo.ap_skewer
		for i = 1, #mo.ap_skewer
			local m = mo.ap_skewer[i]
			if not m or not m.valid continue end	-- will be emptied later

			local useang = mo.angle
			if mo.ap_hsaveangle ~= nil
				useang = mo.ap_hsaveangle
			end

			local x = mo.x + FixedMul(mo.scale*16, cos(useang))
			local y = mo.y + FixedMul(mo.scale*16, sin(useang))

			P_TeleportMove(m, x, y, mo.z)
			m.flags = MF_ENEMY|MF_NOCLIP	-- force

			if mo.ap_handstate ~= HS_DASH	-- wtf!
				--print("Unexpected state change")
				m.flags = m.info.flags | MF_SHOOTABLE
				AP_sliceBadnik(m, mo)
				P_KillMobj(m, mo, mo)
			end

		end

		if mo.ap_handstate ~= HS_DASH
			mo.ap_skewer = nil	-- foh
		end
	end

	if mo.ap_handstate == HS_FREEFALL
	and P_IsObjectOnGround(mo)
		mo.ap_handstate = HS_NONE	-- remove freefall
	end

	local spinbutt = BT_SPIN or BT_USE
	if p.cmd.buttons & spinbutt
		mo.ap_spinbuf = $ and $+1 or 1
	else
		mo.ap_spinbuf = 0
	end

	if p.ap_forcestrafe
	and mo.ap_handstate ~= HS_DIVE
		p.pflags = $ & ~PF_FORCESTRAFE	-- remove forcestrafe
		p.ap_forcestrafe = nil
	end

	if p.ap_analogbuffer
	and not mo.ap_handstate
		p.pflags = $|PF_ANALOGMODE
		p.ap_analogbuffer = nil
	end

	if mo.ap_handstate == HS_DONE
		if P_IsObjectOnGround(mo)
			mo.ap_wasonground = true
			if mo.ap_hsjumpbuf

				if p.charability == CA_ANGELWINGS
					mo.ap_wingstate = HS_NONE
				end
				-- Jump buffer.
				-- Don't forget to take away the wingstate so that we don't end up accidentally floating instead of jumping.
				P_DoJump(p)
				mo.ap_hsjumpbuf = nil
			end
		end

		local spinbutt = BT_SPIN or BT_USE	-- yes this is annoying to take into account :V
		if not (p.cmd.buttons & spinbutt)
		and mo.ap_wasonground	-- make sure we don't get multiple mid air uses
			mo.ap_handstate = HS_NONE	-- don't let us hold the ability
			mo.ap_wasonground = nil
		end
	end

	if mo.ap_handstate == HS_DASH
		mo.state = S_PLAY_HANDSONIC

		if mo.ap_overdrive	-- in overdrive, you can freely control hand sonic!

			if p.pflags & PF_ANALOGMODE	-- ugh, use a real control scheme, please...
				p.ap_analogbuffer = 1
				p.pflags = $ & ~PF_ANALOGMODE
			end

			p.ap_forcestrafe = true
			p.pflags = $ |PF_FORCESTRAFE

			mo.angle= $ - p.cmd.sidemove*ANG1/8	-- allow control via strafe

			P_InstaThrust(mo, mo.angle, mo.ap_hsavespeed)	-- yep this gets dumb
			mo.momz = 0	-- cancel momz
			p.drawangle = mo.angle
			mo.ap_hsaveangle = mo.angle	-- update this if overdrive runs out
			AP_advanceGhost(mo)

			if p.cmd.buttons & spinbutt
			and mo.ap_handtimer < 2
				mo.ap_handtimer = $+1	-- stall the timer so long as you keep going!
			end
		else
			mo.angle = mo.ap_hsaveangle
			p.drawangle = mo.ap_hsaveangle
			--mo.height = skins[mo.skin].height/2
			P_InstaThrust(mo, mo.ap_hsaveangle, mo.ap_hsavespeed)
			mo.momz = 0	-- cancel momz
			AP_advanceGhost(mo)
		end

		if not mo.ap_handtimer

			-- scan for enemies and perform a reach slash if any is detected:
			AP_handSonicSlash(mo)
			AP_hresetStats(p)

			mo.ap_handstate = HS_DONE
			mo.state = P_IsObjectOnGround(mo) and S_PLAY_WALK or S_PLAY_MIDFALL
			if not P_IsObjectOnGround(mo)
				p.pflags = $|PF_JUMPED
				mo.ap_handstate = HS_FREEFALL
				if p.charability == CA_ANGELWINGS
					mo.ap_wingstate = WS_FREEFALL	-- freefall if you use this ability first
				end
			end
		end
		mo.ap_handtimer = $-1
		AP_FOFBuster(mo)

		-- jump buffer:
		if p.cmd.buttons & BT_JUMP
			mo.ap_hsjumpbuf = true
		end
	elseif mo.ap_handstate == HS_DIVE

		if p.pflags & PF_ANALOGMODE	-- ugh, use a real control scheme, please...
			p.ap_analogbuffer = 1
			p.pflags = $ & ~PF_ANALOGMODE
		end

		p.ap_forcestrafe = (p.pflags & PF_FORCESTRAFE) +1	-- in case this was somehow already applied?
		p.pflags = $|PF_FORCESTRAFE	-- yep. that's me.
		mo.ap_hsaveangle = $ - p.cmd.sidemove*ANG1/32	-- allow some slight control

		AP_advanceGhost(mo)
		mo.angle = mo.ap_hsaveangle
		p.drawangle = mo.ap_hsaveangle
		mo.state = S_PLAY_HANDSONIC
		--mo.height = skins[mo.skin].height/2	-- THIS BREAKS IN REVERSE GRAVITY. HOW.
		local a = ANG1*45
		local fspd = FixedMul(cos(a), mo.ap_hsavespeed)
		P_InstaThrust(mo, mo.ap_hsaveangle, fspd)
		mo.momz = -(fspd/2)*P_MobjFlip(mo)
		mo.ap_hsavespeed = $ + mo.scale*3

		mo.ap_handtimer = $+1
		if mo.ap_handtimer == TICRATE	-- double speed after a bit
			S_StartSound(mo, sfx_bkpoof)
			mo.ap_hsavespeed = $*2	-- lol
		end

		AP_FOFBuster(mo)

		if P_IsObjectOnGround(mo)	-- hit the floor, cause a nuke
			AP_hresetStats(p)
			if p.charability == CA_ANGELWINGS
				AP_despawnWings(p)
			end
			mo.ap_handstate = HS_DONE

			if mo.ap_handtimer < TICRATE
				for i = 1, 4
					AP_handSonicSlash(mo, mo.scale*320)	-- this is a disgusting hack but it's funny.
					mo.angle = $+ ANG1*90
				end
				P_NukeEnemies(mo, mo, 320*FRACUNIT)	-- for visuals.
				S_StartSound(mo, sfx_s3k96)
			else
				P_NukeEnemies(mo, mo, 1536*FRACUNIT)
				P_StartQuake(FRACUNIT*80, 8)
				P_FlashPal(p, 1, 2)
				S_StartSound(mo, sfx_s3k9b)
			end
		elseif mo.eflags & MFE_GOOWATER	-- goo water never ceases to be a pain in the ass

			AP_hresetStats(p)
			if p.charability == CA_ANGELWINGS
				AP_despawnWings(p)
			end
			mo.ap_handstate = HS_DONE
			mo.state = S_PLAY_MIDFALL
		end
	end
end)

-- skew m to mo
local function AP_skewer(m, mo)
	if m.flags & MF_BOXICON
		return	-- no
	end
	m.flags = MF_NOCLIP|MF_ENEMY
	mo.ap_skewer = $ or {}
	mo.ap_skewer[#mo.ap_skewer+1] = m

	server.ap_skewed_pointers = $ or {}	-- the server needs to keep track of skewed enemies incase we leave midgame.
	server.ap_skewed_pointers[#server.ap_skewed_pointers+1] = m

	m.skewed = mo	-- keep track of who skewed us

	S_StartSound(mo, sfx_skew)	-- skewer!
	--mo.ap_handtimer = TICRATE/4	-- reset handsonic timer

	local stab = P_SpawnMobjFromMobj(m, 0, 0, 0, MT_THOK)
	stab.momx = mo.momx/4
	stab.momy = mo.momy/4
	stab.color = mo.color
	stab.scale = mo.scale
	stab.state = S_HANDSONIC_STAB
end

-- can damage hook: Hand Sonic does damage when active
addHook("PlayerCanDamage", function(p, m)

	local mo = p.mo
	if mo.ap_handstate and mo.ap_handstate == HS_DIVE or mo.ap_handstate == HS_DASH

		-- Skewers can be performed on:
		-- predefined objects, non boss enemies with 1 HP and montors

		if (AP_slicedBadniks[m.type] or m.flags & (MF_ENEMY|MF_BOSS) == MF_ENEMY)		-- perform a "skewer" instead
		and m.health == 1				-- last hit
		and mo.ap_handstate == HS_DASH	-- dive won't skewer
			AP_skewer(m, mo)
			return false
		end

		return true
	end
end)

-- Hand Sonic will make you bump off of walls:
addHook("MobjMoveBlocked", function(mo)
	local p = mo.player
	if not p return end	-- bruh
	if mo and mo.ap_handstate == HS_DIVE or mo.ap_handstate == HS_DASH

		P_InstaThrust(mo, mo.ap_hsaveangle, -10*mo.scale)
		P_SetObjectMomZ(mo, FRACUNIT*6)
		mo.state = S_PLAY_SPRING
		mo.angle = mo.ap_hsaveangle
		p.drawangle = mo.ap_hsaveangle	-- don't forget.
		AP_hresetStats(p)

		S_StartSound(mo, sfx_cdfm28)

		/*if mo.ap_handstate == HS_DIVE
			local moms = {mo.momx, mo.momy}
			P_DoPlayerPain(p)
			mo.momx = moms[1]
			mo.momy = moms[2]
			p.drawangle = $ + ANG1*180
		end*/

		for i = 1, 32
			local a = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
			a.state = S_MINECARTSPARK
			a.scale = mo.scale

			a.momx = P_RandomRange(-64, 64)*mo.scale
			a.momy = P_RandomRange(-64, 64)*mo.scale
			a.momz = P_RandomRange(-64, 64)*mo.scale
			a.fuse = 6
		end

		mo.ap_handstate = mo.ap_handstate == HS_DIVE and HS_DONE or HS_FREEFALL

		if p.charability == CA_ANGELWINGS
			mo.ap_wingstate = WS_FREEFALL	--WS_NONE
			AP_despawnWings(p)
		end
	end
end, MT_PLAYER)


-- stop HandSonic if we get damaged
local function AP_handSonicDamage(mo)
	local p = mo.player
	AP_hresetStats(p)
	mo.ap_handstate = HS_NONE
end

addHook("MobjDamage", AP_handSonicDamage, MT_PLAYER)
addHook("MobjDeath", AP_handSonicDamage, MT_PLAYER)

-- Hand Sonic beats spikes!
local function AP_handSonicSpikeCollide(mo, spike)
	-- yes we need to check for z height lmfao...

	if spike.type ~= MT_SPIKE and spike.type ~= MT_WALLSPIKE return end

	if mo.z + mo.height >= spike.z
	and mo.z <= spike.z + spike.height

		if mo and mo.ap_handstate and (mo.ap_handstate == HS_DASH or mo.ap_handstate == HS_DIVE)
			P_KillMobj(spike, mo, mo)
			return false	-- didn't collide
		end
	end
end

addHook("MobjCollide", AP_handSonicSpikeCollide, MT_PLAYER)
addHook("MobjMoveCollide", AP_handSonicSpikeCollide, MT_PLAYER)

-- Don't bounce off of monitors.
local function AP_handSonicTVCollide(mo, tv)
	-- yes we need to check for z height lmfao...

	if not (tv.flags & MF_MONITOR and tv.health) return end

	if mo.z + mo.height >= tv.z
	and mo.z <= tv.z + tv.height
		if mo and mo.ap_handstate and (mo.ap_handstate == HS_DASH or mo.ap_handstate == HS_DIVE)

			if tv.flags & MF_GRENADEBOUNCE	-- gold tv. i stg this flag is ass LOL
				P_DamageMobj(tv, mo, mo)	-- deal normal damage
			else
				mo.ap_ogposition = {tv.x, tv.y, tv.z}	-- keep that in mind for multiplayer respawning items
				AP_skewer(tv, mo)
			end
			return false	-- didn't collide
		end
	end
end

addHook("MobjCollide", AP_handSonicTVCollide, MT_PLAYER)
addHook("MobjMoveCollide", AP_handSonicTVCollide, MT_PLAYER)

-- Touching a spring ends the ability:
local function AP_handSonicSpringCollide(mo, spring)
	-- yes we need to check for z height lmfao...
	local p = mo.player
	if not (spring.flags & MF_SPRING) return end

	if mo.z + mo.height >= spring.z
	and mo.z <= spring.z + spring.height

		if mo and mo.ap_handstate and (mo.ap_handstate == HS_DASH or mo.ap_handstate == HS_DIVE)
			mo.ap_handstate = HS_DONE
			mo.ap_wasonground = true	-- cheat
			AP_hresetStats(p)

			if p.charability == CA_ANGELWINGS
				mo.ap_wingstate = WS_NONE
				AP_despawnWings(p)
			end
		end
	end
end

addHook("MobjCollide", AP_handSonicSpringCollide, MT_PLAYER)
addHook("MobjMoveCollide", AP_handSonicSpringCollide, MT_PLAYER)

local function AP_handSonicPlayerCollide(mo, mo2)
	if mo and mo.valid and mo2 and mo2.valid
	and mo.ap_handstate
	and (mo.ap_handstate == HS_DASH or mo.ap_handstate == HS_DIVE)	-- only while dashing/diving obv
	and mo2.player
	and mo.z + mo.height >= mo2.z
	and mo.z <= mo2.z + mo2.height
	and mo.health
	and not AP_friendlyPlayer(mo.player, mo2.player)
		P_DamageMobj(mo2, mo, mo, nil, DMG_SPIKE)
	end
end

addHook("MobjCollide", AP_handSonicPlayerCollide, MT_PLAYER)
addHook("MobjMoveCollide", AP_handSonicPlayerCollide, MT_PLAYER)

-- shoulddamage: anything that belongs to an enemy and whatnot shouldn't damage us:
addHook("ShouldDamage", function(mo, hit)

	if hit and hit.valid	-- first of all...

		-- most important, we can't get damaged by anything we skewed
		if hit.skewed == mo
			return false
		end

		if mo and mo.ap_handstate and (mo.ap_handstate == HS_DASH or mo.ap_handstate == HS_DIVE)
			if hit.tracer and hit.tracer.valid and hit.tracer.flags & MF_ENEMY|MF_BOSS	--!
			or hit.flags & MF_MISSILE
			or hit.target and hit.target.valid and hit.target.flags & MF_ENEMY|MF_BOSS	--!
			or hit.flags & MF_PAIN	-- god.

				if hit.flags & MF_PAIN
					AP_killPainMobj(hit)
				end

				return false	-- nope
			end
		end
	end
end)

-- Handsonic prevents shield usage:
addHook("ShieldSpecial", function(p)
	if p.charability2 == CA2_HANDSONIC
		return true
	end
end)

-- Handsonic hurtmsg

addHook("HurtMsg", function(p, inf)
	if inf and inf.valid and inf.ap_handstate and (inf.ap_handstate == HS_DASH or inf.ap_handstate == HS_DIVE)

		if not p.mo.health
			print(AP_teamName(p).." was impaled and killed by "..AP_teamName(inf.player).."'s Hand Sonic.")
			return true
		else
			print(AP_teamName(p).." was hurt by "..AP_teamName(inf.player).."'s Hand Sonic.")
			return true
		end
	end
end, MT_PLAYER)

-- Server handling of skewed enemies if a player leaves:
rawset(_G, "AP_serverSkewerHandler", function()
	if not netgame or not server.ap_skewer_pointers or not #server.ap_skewer_pointers return end

	local i = #server.ap_skewer_pointers
	while i do
		local m = server.ap_skewer_pointers[i]

		if not m or not m.valid
			table.remove(server.ap_skewer_pointers, i)
			i = $-1
			continue
		end

		-- put whatever was here out of its misery if a player left
		if not m.skewed or not m.skewed.valid
			m.flags = m.info.flags | MF_SHOOTABLE
			P_KillMobj(m)
			table.remove(server.ap_skewer_pointers, i)
			i = $-1
		end
	end
end)