-- AngelPlayer: Overdrive
-- Enables temporary active ability enhancement once all 7 Chaos Emeralds have been gathered.
-- Overdrive charges by killing enemies and collecting rings
-- Once activated, overdrive lasts for 20s. Timer can be extended by 4s for each enemy killed while active.

--local dbg_overdrive = 1

rawset(_G, "AP_canOverdrive", function(p)

	if not p.mo or p.mo.skin ~= "kanade" or p.powers[pw_carry] == CR_NIGHTSMODE return end

	if dbg_overdrive
		return 1
	end

	return All7Emeralds(emeralds)
end)


-- 7 emerald anim to show overdrive being activated:
local function AP_emeraldAnim(p)

	local mo = p.mo

	if not p.ap_allemeralds
	and AP_canOverdrive(p)
		mo.ap_emeraldtimer = 1
		p.ap_allemeralds = 1	-- set this flag to true.
	end

	if mo.ap_emeraldtimer

		if mo.ap_emeraldtimer == 1
			mo.ap_emeralds = {}
			mo.ap_rotspeed = 1
			mo.ap_emang = 0
			mo.ap_emdist = 512

			for i = 1, 7
				mo.ap_emeralds[i] = P_SpawnMobj(0, 0, 0, MT_THOK)
				mo.ap_emeralds[i].state = mobjinfo[MT_EMERALD1+(i-1)].spawnstate
			end
		end

		local ang = mo.ap_emang
		for i = 1, 7
			local x = mo.x + mo.ap_emdist*cos(ang)
			local y = mo.y + mo.ap_emdist*sin(ang)
			local z = mo.z + mo.height/2 + 16*cos(leveltime*ANG1*2 + (360/7)*ANG1*i)	-- funny z axis variation

			P_TeleportMove(mo.ap_emeralds[i], x, y, z)
			ang = $ + (360/7)*ANG1
		end

		mo.ap_emang = $+ mo.ap_rotspeed*ANG1
		mo.ap_rotspeed = min(32, $+1)
		mo.ap_emdist = max(64, $-48)

		mo.ap_emeraldtimer = $+1

		if mo.ap_emeraldtimer >= TICRATE*2
			S_StartSound(nil, sfx_s3kb3, p)
			P_FlashPal(p, 1, 2)
			for i = 1, 7
				if mo.ap_emeralds[i] and mo.ap_emeralds[i].valid
					P_RemoveMobj(mo.ap_emeralds[i])
				end
			end

			mo.ap_emeralds = nil
			mo.ap_emeraldtimer = nil
		end

		local colours = {SKINCOLOR_EMERALD, SKINCOLOR_PURPLE, SKINCOLOR_BLUE, SKINCOLOR_CYAN, SKINCOLOR_ORANGE, SKINCOLOR_RED, SKINCOLOR_GREY}

		local radius = 32

		local s = P_SpawnMobj(mo.x + P_RandomRange(-radius, radius)*mo.scale, mo.y + P_RandomRange(-radius, radius)*mo.scale, mo.z + P_RandomRange(0, radius)*mo.scale, MT_THOK)
		s.state = S_WINGSPARK1
		s.scale = mo.scale*2
		s.momz = P_RandomRange(1, 4)*mo.scale
		s.color = colours[P_RandomRange(1, #colours)]
	end
end

rawset(_G, "AP_overdrive", function(p)

	local mo = p.mo

	mo.ap_alivetime = $ or 0
	if AP_canOverdrive(p)
		mo.ap_alivetime = $+1	-- for overdrive bar visuals.
	end

	mo.ap_overdrivecharge = $ or 0
	mo.ap_overdrive = $ or 0

	AP_emeraldAnim(p)	-- visuals

	if mo.ap_overdrivecharge >= AP_OVERDRIVE_MAXCHARGE
	and not p.ap_hsavestats
	and not p.ap_savestats	-- red flags
		if p.cmd.buttons & BT_TOSSFLAG
			mo.ap_overdrive = AP_OVERDRIVE_DURATION*TICRATE
			mo.ap_overdrivecharge = 0

			p.ap_overdrivesavestats = p.normalspeed
			p.normalspeed = $ * 133/100

			for i = 1, 32
				local s = P_SpawnMobj(mo.x, mo.y, mo.z + mo.height/2, MT_THOK)
				s.state = S_WINGSPARK1
				s.scale = mo.scale*2
				s.momx = P_RandomRange(-32, 32)*mo.scale
				s.momy = P_RandomRange(-32, 32)*mo.scale
				s.momz = P_RandomRange(-32, 32)*mo.scale
				s.color = mo.color
			end

			AP_wingDust(mo, 32)
			-- sfx_kc4d: Alternative
			S_StartSound(mo, sfx_cdfm40)

			local g = P_SpawnGhostMobj(mo)
			g.destscale = mo.scale*8
			g.scalespeed = mo.scale/4
			g.colorized = true

			g = P_SpawnGhostMobj(mo)
			g.destscale = mo.scale*8
			g.scalespeed = mo.scale/12
			g.colorized = true

		end
	end

	if mo.ap_overdrive
		mo.ap_overdrive = $-1

		if leveltime%6 < 3
			mo.colorized = true
			mo.frame = $|FF_FULLBRIGHT
		else
			mo.colorized = false
		end

		if not AP_canOverdrive(p)
			mo.ap_overdrive = 0	-- :how:
		end

		if not mo.ap_overdrive
			mo.colorized = false
			S_StartSound(mo, sfx_kc65)

			local g = P_SpawnGhostMobj(mo)
			g.destscale = mo.scale*8
			g.scalespeed = mo.scale/4
			g.colorized = true

			g = P_SpawnGhostMobj(mo)
			g.destscale = mo.scale*8
			g.scalespeed = mo.scale/12
			g.colorized = true
		end

		-- visuals:
		local radius = 40

		local s = P_SpawnMobj(mo.x + P_RandomRange(-radius, radius)*mo.scale, mo.y + P_RandomRange(-radius, radius)*mo.scale, mo.z + P_RandomRange(0, radius)*mo.scale, MT_THOK)
		s.state = S_WINGSPARK1
		s.scale = mo.scale*2
		s.momz = P_RandomRange(1, 4)*mo.scale
		s.color = mo.color

	elseif p.ap_overdrivesavestats
		if not p.ap_hsavestats
		and not p.ap_savestats
			p.normalspeed = p.ap_overdrivesavestats	-- it's safe to reset this now
		end
	end
end)

-- fill overdrive gauge by collecting rings: (1 ring = 1%)
-- MT_FLINGRING doesn't fill the gauge.
addHook("TouchSpecial", function(mo, toucher)
	if AP_canOverdrive(toucher.player)
	and not toucher.ap_overdrive
	and mo and mo.valid
	and mo.health
		toucher.ap_overdrivecharge = min(AP_OVERDRIVE_MAXCHARGE, $+1)
	end
end, MT_RING)

rawset(_G, "AP_addOverdrive", function(p)
	if not AP_canOverdrive(p) return end
	if not p.mo or not p.mo.valid return end
	if p.mo.ap_overdrive

		-- in that case, add 4 seconds back.
		p.mo.ap_overdrive = min(AP_OVERDRIVE_DURATION*TICRATE, $ + TICRATE*4)
		return
	end

	p.mo.ap_overdrivecharge = min(AP_OVERDRIVE_MAXCHARGE, $ + AP_OVERDRIVE_KILLCHARGE)
end)

-- you can get knocked back during overdrive but you won't actually lose rings:
addHook("MobjDamage", function(mo, inflictor, source)
	if mo.ap_overdrive
		P_DoPlayerPain(mo.player, source, inflictor)
		S_StartSound(mo, sfx_spkdth)
		return true
	end
end, MT_PLAYER)


-- gets bar offset depending on status
local function AP_getBarOffset(p)

	local y = 0
	local didf12
	-- spectator mode: skipped because the bar wouldn't be drawn there anyway
	if circuitmap
		y = $+8	-- for lap / finished text
	end

	if gametype != GT_COOP and (p.exiting or (G_GametypeUsesLives() and p.lives <= 0))
		if not splitscreen and not didf12
			y = $+8	-- f12 text
			didf12 = true
		end

	elseif ((gametyperules & GTR_RESPAWNDELAY) and p.playerstate == PST_DEAD and p.lives)
		y = $+8	-- respawn text

	end	-- again, spectator stuff is skipped

	if p.charability == 69	-- yes, this is funny
		y = $ + 10
	end

	if (gametype == GT_COOP and (not p.spectator or (not (maptol & TOL_NIGHTS) and G_IsSpecialStage(gamemap))) and (p.exiting or (p.pflags & PF_FINISHED)))

		local numneeded = (G_IsSpecialStage(gamemap) and 4 or CV_FindVar("playersforexit").value)
		if numneeded
			local i
			local total = 0
			local exiting = 0

			for p2 in players.iterate do
				if p2.spectator or not p2.lives continue end

				total = $+1
				if p2.exiting or p2.pflags & PF_FINISHED
					exiting = $+1
				end
			end

			if (numneeded != 4)
				total = $ * CV_FindVar("playersforexit").value
				if (total & 3)
					total = $ + 4 // round up
				end
				total = $/4
			end

			if (exiting < total)
				if (not splitscreen and not didf12)
					y = $+8	-- viewpoint text
					didf12 = true
				end
				total = $ -exiting
				y = $+8	-- players remaining
			end
		end

	elseif ((gametyperules & GTR_TAG) and not p.spectator)
		if (leveltime < CV_FindVar("hidetime").value * TICRATE)
			if (p.pflags & PF_TAGIT)
				if (gametyperules & GTR_BLINDFOLDED)
					y = $+8	-- blindfold
				end
				y = $+8	-- waiting for players to hide...

			else
				y = $+8	-- hide before time runs out
			end

		elseif (gametype == GT_HIDEANDSEEK and not (p.pflags & PF_TAGIT))
			if not splitscreen and not didf12
				y = $+8	-- viewpoint
				didf12 = true
			end

			if (gametyperules & GTR_HIDEFROZEN)
				y = $+8	-- can't move
			end
		end
	end
	-- god this was a huge fucking pain wasn't it?
	return y
end

-- silv_drawBarHUD: for mod support if rings hud is disabled
-- player_t p:	player we draw it for
-- int basex:	base x coord
-- int basey:	base y coord
-- int mbarlen:	max length of the bar
-- int flags:	flags to use

rawset(_G, "AP_overdriveHUD", function(v, p, basex, basey, mbarlen, flags)

	local mo = p.mo	-- just in case

	local t = netgame and leveltime or p.mo.ap_alivetime

	if t < TICRATE return end	-- no bar.

	local maxbarlen = min(mbarlen, (t-TICRATE)*6)


	if mo.skin ~= "kanade" then return end

	for i = 0, 1

		-- draw bar background

		local dx = basex-i
		local dy = basey+i
		local splitfac = splitscreen and 2 or 1

		local backcolor = 31
		v.drawFill(dx, dy, maxbarlen+1, 2*splitfac, backcolor|flags)

		-- fill the bar
		local barcolor
		local barlen
		local msg


		if mo.ap_overdrive
			barcolor = 1
			barlen = mo.ap_overdrive * maxbarlen / (AP_OVERDRIVE_DURATION*TICRATE)

			if i == 1
				msg = "\x82"..G_TicsToSeconds(mo.ap_overdrive).."."..string.format("%02d", G_TicsToCentiseconds(mo.ap_overdrive))
			end

		else
			barcolor = mo.ap_overdrivecharge < AP_OVERDRIVE_MAXCHARGE and 35 or 0
			barlen = mo.ap_overdrivecharge*maxbarlen / AP_OVERDRIVE_MAXCHARGE	-- 100% charge

			if mo.ap_overdrivecharge >= AP_OVERDRIVE_MAXCHARGE
			and i == 1
				msg = "\x82TOSSFLAG: OVERDRIVE"
			end
		end

		-- draw the bar
		if barlen
			v.drawFill(dx, dy, barlen, 1*splitfac, barcolor|flags)
		end

		if i
		and msg
			v.drawString(dx + maxbarlen, dy, msg, flags, "thin-right")
		end
	end
end)

-- Overdrive hud, yes this is mostly recycled from Silver.
hud.add(function(v, p, c)

	if not AP_canOverdrive(p) or not hud.enabled("rings") or p.powers[pw_carry] == CR_NIGHTSMODE return end

	-- bar smooth appearance:
	-- hud doesn't appear until leveltime > ticrate
	local rpatch = v.cachePatch("STTRINGS")
	local h = rpatch.height
	h = $+ AP_getBarOffset(p)
	AP_overdriveHUD(v, p, hudinfo[HUD_RINGS].x +1, hudinfo[HUD_RINGS].y + h + 5, 100, V_PERPLAYER|V_SNAPTOTOP|V_SNAPTOLEFT|V_HUDTRANS)
end)